package tests;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URLClassLoader;

import AST.Program;

public abstract class TestsContainer {
	protected abstract String getTestDir();
	protected abstract String getTestPackage();
	
	protected void runPositiveTestAndAssertEquals(int testCase, String errorMessageOnCompilerError,
			String errorMessageOnAssertError, Object expectedValue) throws Exception {
		runPositiveTestAndAssertEquals(toZeroPaddedString(testCase), errorMessageOnCompilerError,
				errorMessageOnAssertError, expectedValue);
	}
	
	protected void runPositiveTestAndAssertEquals(int testCase, String errorMessageOnCompilerError,
			String messageOnStdOut, String messageOnStdErr,
			String errorMessageOnAssertError, Object expectedValue) throws Exception {
		runPositiveTestAndAssertEquals(toZeroPaddedString(testCase), errorMessageOnCompilerError,
				messageOnStdOut, messageOnStdErr, errorMessageOnAssertError, expectedValue);
	}
	
	protected void runPositiveTestAndAssertEquals(String testCase, String errorMessageOnCompilerError,
			String errorMessageOnAssertError, Object expectedValue) throws Exception {
		Object result = runPositiveTest(testCase, errorMessageOnCompilerError);
		assertEquals(errorMessageOnAssertError, expectedValue, result);
	}
	
	protected void runPositiveTestAndAssertEquals(String testCase, String errorMessageOnCompilerError,
			String messageOnStdOut, String messageOnStdErr,
			String errorMessageOnAssertError, Object expectedValue) throws Exception {
		Object result = runPositiveTest(testCase, errorMessageOnCompilerError, messageOnStdOut, messageOnStdErr);
		assertEquals(errorMessageOnAssertError, expectedValue, result);
	}
	
	protected Object runPositiveTest(int testCase, String errorMessageOnCompilerError) throws Exception {
		return runPositiveTest(toZeroPaddedString(testCase), errorMessageOnCompilerError, "", "");
	}
	
	protected Object runPositiveTest(int testCase, String errorMessageOnCompilerError,
			String messageOnStdOut, String messageOnStdErr) throws Exception {
		return runPositiveTest(toZeroPaddedString(testCase), errorMessageOnCompilerError,
				messageOnStdOut, messageOnStdErr);
	}
	
	protected Object runPositiveTest(String tc, String errorMessageOnCompilerError) throws Exception {
		return runPositiveTest(tc, errorMessageOnCompilerError, "", "");
	}
	
	protected Object runPositiveTest(String tc, String errorMessageOnCompilerError,
			String messageOnStdOut, String messageOnStdErr)
			throws Exception {
		// Redirect streams, test, restore streams
		StdStreamRedirection ssr = new StdStreamRedirection().redirect();
		Object result = null;
		
		try {
			Program program = CompileHelper.compile(getTestDir()+"test"+tc+"/Test"+tc+".jaaa");
			if(program != null) {
				program.transformation();
				program.generateClassfile();
				result = CompileHelper.invokeStatic(getTestClassName(tc), "test", new Class[0], new Object[0]);
			}
			assertEquals("Wrong output on stdOut", messageOnStdOut, ssr.out.byteStream.toString());
			assertEquals("Wrong output on stdErr", messageOnStdErr, ssr.err.byteStream.toString());
		} finally {
			ssr.restore();
		}
		
		return result;
	}
	
	protected void runNegativeTest(int testCase, String errorMessageOnCompilerError) throws Exception {
		runNegativeTest(toZeroPaddedString(testCase), errorMessageOnCompilerError, "");
	}
	
	protected void runNegativeTest(String tc, String errorMessageOnCompilerError) throws Exception {
		runNegativeTest(tc, errorMessageOnCompilerError, "");
	}
	
	protected void runNegativeTest(int testCase, String errorMessageOnCompilerError, String messageOnStdOut)
		throws Exception {
		runNegativeTest(toZeroPaddedString(testCase), errorMessageOnCompilerError, messageOnStdOut);
	}
	
	protected void runNegativeTest(String testCase, String errorMessageOnCompilerError, String messageOnStdOut)
		throws Exception {
		// Redirect streams, test, restore streams
		StdStreamRedirection ssr = new StdStreamRedirection().redirect();
		
		try {
			Program program = CompileHelper.compile(getTestDir()+"test"+testCase+"/Test"+testCase+".jaaa");
			assertNull(errorMessageOnCompilerError, program);
			assertEquals("Wrong output on stdOut", messageOnStdOut, ssr.out.byteStream.toString());
			assertEquals("Wrong output on stdErr", "", ssr.err.byteStream.toString());
		} finally {
			ssr.restore();
		}
	}
	
	private String getTestClassName(String tc) {
		return "tests."+getTestPackage()+".test"+tc+".Test"+tc;
	}
	
	public URLClassLoader getTestClassLoader()
			throws MalformedURLException, IOException {
		return CompileHelper.getClassLoader();
	}
	
	public Object newInstanceOfTestClass(int testCase)
		throws MalformedURLException, InstantiationException,
			IllegalAccessException, ClassNotFoundException, IOException {
		return newInstanceOfTestClass(toZeroPaddedString(testCase));
	}
	
	public Object newInstanceOfTestClass(String tc)
		throws MalformedURLException, InstantiationException,
			IllegalAccessException, ClassNotFoundException, IOException {
		return CompileHelper.newInstance(getTestClassName(tc));
	}
	
	private String toZeroPaddedString(int testCase) {
		return String.format("%1$02d", testCase);
	}
	
	private static class StdStreamRedirection {
		final StreamRedirection out;
		final StreamRedirection err;
		
		public StdStreamRedirection() {
			out = new StreamRedirection(System.out);
			err = new StreamRedirection(System.err);
		}
		
		public StdStreamRedirection redirect() {
			System.setOut(out.newPS);
			System.setErr(err.newPS);
			return this;
		}
		
		public void restore() {
			System.setOut(out.oldPS);
			System.setErr(err.oldPS);
		}
	}
	
	private static class StreamRedirection {
		
		private PrintStream oldPS;
		private ByteArrayOutputStream byteStream;
		private PrintStream newPS;
		
		public StreamRedirection(PrintStream oldPS) {
			this.oldPS = oldPS;
			byteStream = new ByteArrayOutputStream();
			newPS = new PrintStream(byteStream);
		}
	}
}
